<?php

declare(strict_types=1);
namespace Drupal\translation_bliss\Drush\Commands\PotxCommands;

use Consolidation\AnnotatedCommand\CommandData;
use Consolidation\AnnotatedCommand\Events\CustomEventAwareInterface;
use Consolidation\AnnotatedCommand\Events\CustomEventAwareTrait;
use Consolidation\AnnotatedCommand\Hooks\HookManager;
use Consolidation\OutputFormatters\StructuredData\RowsOfFields;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\translation_bliss\Drush\Commands\PotxCommands\Utility\PotxResult;
use Drupal\translation_bliss\Drush\Commands\PotxCommands\Utility\TranslationStatusFilter;
use Drupal\translation_bliss\Drush\Commands\Utility\TmpCwd;
use Drush\Attributes as CLI;
use Drush\Commands\AutowireTrait;
use Drush\Commands\DrushCommands;
use Symfony\Component\Console\Input\InputOption;

/**
 * Class PotxXCommands.
 *
 * Potx commands for site builders.
 *
 * @internal
 */
class TranslationBlissPotxCommands extends DrushCommands implements CustomEventAwareInterface {

  use CustomEventAwareTrait;
  use AutowireTrait;

  public const PotxxCommandName = 'bliss:potx';

  // Do not use POTX_API_CURRENT, it breaks drush discovery.
  public const PotxApi = 8;
  public const PotxSaveCallback = '_potx_save_string';
  public const PotxVersionCallback = '_potx_save_version';


  const DrushHookLocalInit = 'potx_bliss_local_init';
  const DrushHookExploreDirs = 'potx_bliss_explore_dirs';
  const DrushHookProcessFiles = 'potx_bliss_process_files';
  const DrushHookFinishProcessing = 'potx_bliss_finish_processing';

  const DefaultSourceDirs = 'modules/custom';

  const DefaultTargetDir = '../translations/custom';

  /**
   * @var list<string>
   */
  protected array $sourceDirs;

  protected string $targetDir;

  public function __construct(
    protected readonly ModuleHandlerInterface $moduleHandler,
    protected readonly LanguageManagerInterface $languageManager,
    protected readonly ConfigFactoryInterface $configFactory,
    protected readonly FileSystemInterface $fileSystem,
  ) {
    parent::__construct();
  }

  #[CLI\Hook(type: HookManager::ARGUMENT_VALIDATOR, target: self::PotxxCommandName)]
  public function validateExtract(CommandData $commandData): void {
    $options = $commandData->options();

    /** @var string[] $sourceDirs */
    $this->sourceDirs = explode(',', $options['source'] ?: static::DefaultSourceDirs);
    foreach ($this->sourceDirs as &$sourceDir) {
      $this->normalizeDir($sourceDir);
      $this->validateDir($sourceDir);
    }

    /** @var string $target */
    $this->targetDir = $options['target'] ?: self::DefaultTargetDir;
    $this->normalizeDir($this->targetDir);
    $this->validateDir($this->targetDir, createIfNeeded: TRUE);
  }

  private function normalizeDir(string &$dir): void {
    $dir = rtrim($dir, '/');
  }

  private function validateDir(string $dir, $createIfNeeded = FALSE): void {
    if ($createIfNeeded && !is_dir($dir)) {
      $this->fileSystem->prepareDirectory($dir, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS);
    }
    if (!is_dir($dir)) {
      $cwd = getcwd();
      throw new \UnexpectedValueException("Not a directory: '$dir', cwd='$cwd'");
    }
  }

  #[CLI\Hook(type: HookManager::ARGUMENT_VALIDATOR, target: TranslationBlissPotxCommands::PotxxCommandName)]
  public function validatePotxInstalled(CommandData $commandData): void {
    // Ensure that potx is installed. Potx depends on locale.
    if (!$this->moduleHandler->moduleExists('potx')) {
      throw new \Exception('Please install the potx module required for this command.');
    }
  }

  /**
   * Extract translatable strings from custom code and all config, optimized for site owners.
   */
  #[CLI\Command(name: self::PotxxCommandName, aliases: ['potxx'])]
  #[CLI\Usage(name: 'potxx', description: 'Translation template extraction optimized for site owners.')]
  #[CLI\Option(name: 'source', description: 'One or more source folders. Defaults to "modules/custom".')]
  #[CLI\Option(name: 'target', description: 'Target folder. Defaults to "../translations/custom".')]
  #[CLI\FieldLabels(labels: [
    'files' => 'Files',
    'strings' => 'Strings',
    'errors' => 'Errors',
  ])]
  public function extractToFiles(array $options = [
    'source' => InputOption::VALUE_REQUIRED,
    'target' => InputOption::VALUE_REQUIRED,
  ]): RowsOfFields {
    // Extract.
    $result = $this->extract($this->sourceDirs);
    $rows[] = [
      'files' => count($result->files),
      'strings' => count($result->strings),
      'errors' => count($result->errors),
    ];

    $renames = [];
    // Build custom.pot
    $this->buildFile($renames, langcode: NULL, filterShipped: FALSE);

    // Build costom.XX.po & shipped.XX._po
    foreach ($this->languageManager->getLanguages() as $language) {
      if ($language->getId() === 'en') {
        if (!$this->configFactory->get('locale.settings')->get('translate_english')) {
          continue;
        }
      }
      $this->buildFile($renames, $language->getId(), filterShipped: FALSE);
      $this->buildFile($renames, $language->getId(), filterShipped: TRUE);
    }

    // Write and rename.
    $this->writeFiles($renames);

    // Print results.
    if (!empty($result->errors)) {
      // Fool potx.
      $logError = [$this->logger(), 'error'];
      $this->output()->writeln(dt('Errors'));
      foreach (array_unique($result->errors) as $error) {
        $logError($error);
      }
    }
    $this->output()->writeln('');
    return new RowsOfFields($rows);
  }

  protected function extract(
    array $sourceDirs,
  ): PotxResult {
    // Include library.
    $this->moduleHandler->loadInclude('potx', 'inc');
    $this->moduleHandler->loadInclude('potx', 'inc', 'potx.local');

    $files = [];

    // Silence error message reporting. Messages will be reported by at the end.
    potx_status('set', POTX_STATUS_SILENT);

    foreach ($sourceDirs as $sourceDir) {
      potx_local_init($sourceDir);
    }
    foreach ($this->getCustomEventHandlers(self::DrushHookLocalInit) as $hookLocalInit) {
      $hookLocalInit($sourceDirs);
    }

    foreach ($sourceDirs as $sourceDir) {
      $files = array_merge($files, _potx_explore_dir($sourceDir, '*', self::PotxApi, TRUE));
    }
    foreach ($this->getCustomEventHandlers(self::DrushHookExploreDirs) as $hookExploreDirs) {
      $hookFiles = $hookExploreDirs($sourceDirs);
      $files = array_merge($files, $hookFiles);
    }

    foreach ($files as $file) {
      $this->output()->writeln("Processing $file...");
      _potx_process_file($file, 0, self::PotxSaveCallback, self::PotxVersionCallback, self::PotxApi);
    }
    foreach ($this->getCustomEventHandlers(self::DrushHookProcessFiles) as $hookProcessFiles) {
      $hookProcessFiles($files);
    }

    potx_finish_processing(self::PotxSaveCallback, self::PotxApi);
    foreach ($this->getCustomEventHandlers(self::DrushHookFinishProcessing) as $hookFinishProcessing) {
      $hookFinishProcessing();
    }

    // Get errors, if any.
    $errors = potx_status('get', TRUE);

    // Get saved strings.
    /** @var string[] $strings */
    $strings = _potx_save_string(NULL, NULL, NULL, 0, POTX_STRING_RUNTIME);

    return new PotxResult($files, $strings, $errors);
  }

  protected function buildFile(
    array &$renames,
    ?string $langcode,
    bool $filterShipped,
  ): void {
    // Potx always appends .pot, so use temporary name.
    // This has the additional benefit of updates being atomic, i.e. that an
    // interrupted build does not change the target file.
    [$temporaryName, $name, $ext] = $langcode ?
      (
      $filterShipped ?
        ["_shipped_$langcode", "shipped.$langcode", '_po'] :
        ["_custom_$langcode", "custom.$langcode", 'po']
      ) :
      ['_all', 'all', 'pot'];
    $renames["$this->targetDir/$temporaryName.pot"] = "$this->targetDir/$name.$ext";

    $statusFilter = new TranslationStatusFilter($langcode, $filterShipped);
    _potx_build_files(
      string_mode: POTX_STRING_RUNTIME,
      build_mode: POTX_BUILD_SINGLE,
      force_name: $temporaryName,
      save_callback: $statusFilter,
      version_callback: '_potx_save_version',
      header_callback: '_potx_get_header',
      template_export_langcode: $langcode,
      translation_export_langcode: $langcode,
      api_version: self::PotxApi
    );

    if ($langcode) {
      global $_potx_store;
      // Remove the source comments from .po files, leave them in .pot file.
      $strings = $_potx_store[$temporaryName]['strings'] ?? NULL;
      if ($strings) {
        $strings = preg_replace('/^#:.*?\n/um', '', $strings);
      }
      $_potx_store[$temporaryName]['strings'] = $strings;

      $header = $_potx_store[$temporaryName]['header'] ?? NULL;
      // Also add a custom / shipped explanation.
      if ($header) {
        if (!$filterShipped) {
          $header = <<<EOD
          # Translation Bliss:
          # Custom strings that are untranslated or marked as custom translated.
          # They may or may not be also used by a shipped module.
          #
          $header
          EOD;
        }
        else {
          $header = <<<EOD
          # Translation Bliss:
          # Custom strings that are also used by a shipped module's config.
          # If translation is changed, this affects custom and shipped usage,
          # and the translation will go to 'custom.xx.po'.
          #
          $header
          EOD;
        }
      }
      $_potx_store[$temporaryName]['header'] = $header;
    }
  }

  protected function writeFiles(
    array $renames,
  ): void {
    $tmpCwd = TmpCwd::create($this->targetDir)
      ?? throw new \RuntimeException("Can not change to dir $this->targetDir");

    _potx_write_files();

    unset($tmpCwd);

    foreach ($renames as $from => $to) {
      if (is_file($from)) {
        rename($from, $to);
      }
      else {
        unlink($to);
      }
    }
  }

}
