<?php

declare(strict_types=1);
namespace Drupal\translation_bliss\Drush\Commands\PotxCommands\Utility;

use Drupal\Core\Database\Connection;

/**
 * Annotates strings with their translation status.
 *
 * @internal
 */
final class TranslationStatusFilter {

  const PREFIX_FOR_CUSTOMIZED = 'CUSTOMIZED';
  const PREFIX_FOR_SHIPPED = 'SHIPPED';
  const PREFIX_FOR_UNTRANSLATED = 'UNTRANSLATED';

  public function __construct(
    private readonly ?string $langcode,
    private readonly bool $filterShipped,
    private ?Connection $connection = NULL,
  ) {
    $this->connection = $this->connection ?? \Drupal::database();
  }

  public function __invoke(...$arguments) {
    return _potx_save_string(...$arguments);
  }

  /**
   * @see \_potx_save_string
   */
  private function _potx_save_string(
    ?string $value = NULL,
    ?int $context = NULL,
    ?string $file = NULL,
    int $line = 0,
    int $string_mode = POTX_STRING_RUNTIME
  ): ?array {
    // Pass through storing strings.
    if (isset($value)) {
      return \_potx_save_string($value, $context, $file, $line, $string_mode);
    }
    // Filter reading of string if requested.
    $strings = \_potx_save_string($value, $context, $file, $line, $string_mode);
    assert(is_array($strings));
    return $this->filter($strings);
  }

  private function filter(array $strings): array {
    if (!$this->langcode) {
      return $strings;
    }
    $filtered = [];
    /** @see \_potx_build_files */
    foreach ($strings as $string => $stringInfo) {
      foreach ($stringInfo as $context => $fileInfo) {
        // https://git.drupalcode.org/project/potx/-/merge_requests/24/diffs
        $target = $this->connection->query(
          query: 'SELECT t.translation, t.customized FROM {locales_source} s '
          . 'LEFT JOIN {locales_target} t ON t.lid = s.lid '
          . 'WHERE s.source = :source AND s.context = :context AND t.language = :langcode',
          args: [
            ':source' => $string,
            ':context' => $context,
            ':langcode' => $this->langcode,
          ]
        )->fetchObject();

        $translationStatus = ($target && $target->translation) ?
          (
          intval($target->customized) ?
            self::PREFIX_FOR_CUSTOMIZED :
            self::PREFIX_FOR_SHIPPED
          ) :
          self::PREFIX_FOR_UNTRANSLATED;


        $willUseString = $this->filterShipped ?
          ($translationStatus === self::PREFIX_FOR_SHIPPED) :
          ($translationStatus !== self::PREFIX_FOR_SHIPPED);

        if ($willUseString) {
          $filtered[$string][$context] = $fileInfo;
        }
      }
    }
    return $filtered;
  }

}
