<?php

declare(strict_types=1);
namespace Drupal\translation_bliss\Drush\Commands\PotxCommandsExtensions;

use Consolidation\AnnotatedCommand\CommandData;
use Consolidation\AnnotatedCommand\Hooks\HookManager;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\TypedConfigManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Site\Settings;
use Drupal\Core\TypedData\TraversableTypedDataInterface;
use Drupal\Core\TypedData\TypedDataInterface;
use Drupal\locale\LocaleDefaultConfigStorage;
use Drupal\translation_bliss\ConfigExclude\TranslationBlissConfigExcludeInterface;
use Drupal\translation_bliss\Drush\Commands\PotxCommands\TranslationBlissPotxCommands;
use Drush\Attributes as CLI;
use Drush\Commands\AutowireTrait;
use Drush\Commands\DrushCommands;

/**
 * @internal
 */
final class TranslationBlissConfigPotxCommands extends DrushCommands {

  use AutowireTrait;

  const DoNotExtractOption = 'no-config';

  protected bool $doNotExtract;

  public function __construct(
    private readonly ConfigFactoryInterface $configFactory,
    private readonly TypedConfigManagerInterface $typedConfigManager,
    private readonly ModuleHandlerInterface $moduleHandler,
    private readonly TranslationBlissConfigExcludeInterface $configExclude,
    private readonly LocaleDefaultConfigStorage $localeDefaultConfigStorage,
  ) {
    parent::__construct();
  }

  private function extractConfig(): void {
    // Parsing config via _potx_process_module_schemas works out badly, as
    // there is no simple way to determine the providing module and load its
    // schema.
    // So load runtime config, but use config/sync directory to hint file.
    $configNames = $this->configFactory->listAll();
    // Prime the static cache.
    $this->configFactory->loadMultiple($configNames);
    foreach ($configNames as $configName) {
      $config = $this->configFactory->getEditable($configName);
      $typedConfig = $this->typedConfigManager->createFromNameAndData($configName, $config->getRawData());
      $this->output()->writeln("Processing config $configName");
      $this->processTypedConfig($typedConfig);
    }
  }

  private function processTypedConfig(TypedDataInterface $element): void {
    // Ignore excluded config like webform elements.
    if ($this->configExclude->isConfigTranslationExcluded($element)) {
      $this->output()->writeln("Skipping excluded config property {$element->getPropertyPath()}");
      return;
    }
    // Ignore shipped config.
    if ($element->getValue() === $this->getLocaleDefaultData($element)) {
      return;
    }

    /** Adapted from @see \Drupal\locale\LocaleConfigManager::getTranslatableData */
    if ($element instanceof TraversableTypedDataInterface) {
      foreach ($element as $property) {
        $this->processTypedConfig($property);
      }
    }
    else {
      $value = $element->getValue();
      $definition = $element->getDataDefinition();
      if (!empty($definition['translatable']) && $value !== '' && $value !== NULL) {
        $context = $definition['translation context'] ?? NULL;
        [$fileName, $mockLine] = $this->getConfigFileAndMockLine($element);
        _potx_save_string($value, $context, $fileName, $mockLine);
      }
    }
  }

  private function getConfigFileAndMockLine(TypedDataInterface $element): array {
    $configName = $element->getRoot()->getPropertyPath();
    $propertyPath = $element->getPropertyPath();
    $configPath = substr($propertyPath, strlen($configName) + 1);

    $dirPath = Settings::get('config_sync_directory');
    return ["$dirPath/$configName.yml", $configPath];
  }

  private function getLocaleDefaultData(TypedDataInterface $element): mixed {
    $configName = $element->getRoot()->getPropertyPath();
    $propertyPath = $element->getPropertyPath();
    $configPath = substr($propertyPath, strlen($configName) + 1);
    $localeDefaultConfig = $this->localeDefaultConfigStorage->read($configName);
    if ($configPath && isset($localeDefaultConfig)) {
      assert(is_array($localeDefaultConfig));
      $parents = explode('.', $configPath);
      $result = NestedArray::getValue($localeDefaultConfig, $parents);
    }
    else {
      $result = $localeDefaultConfig;
    }
    return $result;
  }


  #[CLI\Hook(type: HookManager::OPTION_HOOK, target: TranslationBlissPotxCommands::PotxxCommandName)]
  #[CLI\Option(name: self::DoNotExtractOption, description: 'Do not extract config/sync values.')]
  public function optionsPotxX($options = [self::DoNotExtractOption => FALSE]): void {}

  #[CLI\Hook(type: HookManager::ARGUMENT_VALIDATOR, target: TranslationBlissPotxCommands::PotxxCommandName)]
  public function readOptions(CommandData $commandData): void {
    $this->doNotExtract = $commandData->input()->getOption(self::DoNotExtractOption) ?? FALSE;
  }

  #[CLI\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookFinishProcessing)]
  public function hookPotxFinishProcessing(): void {
    if ($this->doNotExtract) {
      $this->output()->writeln('Skipping config extraction.');
      return;
    }
    $this->output()->writeln('Extracting config.');
    $this->extractConfig();
  }

  /**
   * @param list<string> $sourceDirs
   */
  #[CLI\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookLocalInit)]
  public function hookLocalInit(array $sourceDirs): void {
    // Nothing to do. Just to show off the API:
  }

  /**
   * @param list<string> $sourceDirs
   *
   * @return list<string>
   *   A list of files to process.
   */
  #[CLI\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookExploreDirs)]
  public function hookExploreDirs(array $sourceDirs): array {
    // Nothing to do. Just to show off the API:
    return $sourceDirs;
  }

  /**
   * @param list<string> $files
   */
  #[CLI\Hook(type: HookManager::ON_EVENT, target: TranslationBlissPotxCommands::DrushHookProcessFiles)]
  public function hookProcessFiles(array $files): void {
    // Nothing to do. Just to show off the API:
  }

}
